/* upstart
 *
 * Copyright © 2008 Canonical Ltd.
 * Author: Scott James Remnant <scott@netsplit.com>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fnmatch.h>

#include <nih/macros.h>
#include <nih/alloc.h>
#include <nih/string.h>
#include <nih/signal.h>
#include <nih/io.h>
#include <nih/main.h>
#include <nih/option.h>
#include <nih/command.h>
#include <nih/logging.h>
#include <nih/error.h>
#include <nih/dbus.h>

#include "com.ubuntu.Upstart.h"
#include "com.ubuntu.Upstart.Job.h"

/* Prototypes for option and command functions */
int env_option          (NihOption *option, const char *arg);
int start_action        (NihCommand *command, char * const *args);
int stop_action         (NihCommand *command, char * const *args);
int status_action       (NihCommand *command, char * const *args);
int list_action         (NihCommand *command, char * const *args);
int emit_action         (NihCommand *command, char * const *args);
int jobs_action         (NihCommand *command, char * const *args);
int version_action      (NihCommand *command, char * const *args);
int log_priority_action (NihCommand *command, char * const *args);

/**
 * no_wait:
 *
 * Whether to wait for a job or event to be finished before exiting or not.
 **/
int no_wait = FALSE;

/**
 * emit_env:
 *
 * Environment variables to emit along with the event.
 **/
char **emit_env = NULL;

/**
 * conn:
 *
 * DBus Connection to init
 **/
DBusConnection *conn = NULL;

/**
 * proxy:
 *
 * Proxy object for the main control interface
 **/
NihDBusProxy *proxy = NULL;

/**
 * env_option:
 * @option: NihOption invoked,
 * @arg: argument to parse.
 *
 * This option setter is used to append @arg to the list of environment
 * variables pointed to by the value member of option, which must be a
 * pointer to a char **.
 *
 * If @arg does not contain an '=', the current value from the environment
 * is taken instead.
 *
 * The arg_name member of @option must not be NULL.
 *
 * Returns: zero on success, non-zero on error.
 **/
int
env_option (NihOption  *option,
	    const char *arg)
{
	char ***value;

	nih_assert (option != NULL);
	nih_assert (option->value != NULL);
	nih_assert (arg != NULL);

	value = (char ***)option->value;

	if (strchr (arg, '=')) {
		NIH_MUST (nih_str_array_add (value, NULL, NULL, arg));
	} else {
		char *env, *new_arg;

		env = getenv (arg);
		if (env) {
			NIH_MUST (new_arg = nih_sprintf (NULL, "%s=%s",
							 arg, env));
			NIH_MUST (nih_str_array_addp (value, NULL, NULL,
						      new_arg));
		}
	}

	return 0;
}


/**
 * start_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "start" command; it takes one or more
 * arguments that give the names, or ids, of jobs that should be started.
 * If no arguments are given, the UPSTART_JOB environment variable is used in
 * preferences.
 *
 * Returns: command exit status.
 **/
int
start_action (NihCommand   *command,
	      char * const *args)
{
	NihDBusProxy *job;
	char *job_path;
	char *inst_path;
	char *us_inst;
	char *job_name = *args;
	char **env = (char**)args + 1;

	if (! job_name) {
		job_name = getenv("UPSTART_JOB");
		us_inst = getenv("UPSTART_INSTANCE");
		if (! job_name) {
			nih_fatal (_("missing job name"));
			return 1;
		}

		if (us_inst) {
			char *env_str;
			NIH_MUST (env_str = nih_sprintf (NULL,
				"UPSTART_INSTANCE=%s", us_inst));
			env = nih_str_array_add (&env, NULL, NULL, us_inst);
		}
	}

	if (control_get_job_by_name (proxy, job_name, &job_path)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to start job: %s"),
			   err->message);

		return 1;
	}

	if (! job_path) {
		nih_fatal (_("No such job: %s"), job_name);
		return 1;
	}

	NIH_MUST (job = nih_dbus_proxy_new (NULL, conn,
			"com.ubuntu.Upstart.Job", job_path));

	if (job_class_start (job, env, &inst_path)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to start job: %s"),
			   err->message);

		return 1;
	}

	return 0;
}

/**
 * stop_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "stop" command; it takes one or more
 * arguments that give the name of a job that should be stopped and environment
 * variables which the job should have. If no arguments are given, the
 * UPSTART_JOB environment variable is used in preferences.
 *
 * Returns: command exit status.
 **/
int
stop_action (NihCommand   *command,
	     char * const *args)
{
	NihDBusProxy *job;
	char *job_path;
	char *inst_path;
	char *us_inst;
	char *job_name = *args;
	char **env = (char**)args + 1;

	if (! job_name) {
		job_name = getenv ("UPSTART_JOB");
		us_inst = getenv ("UPSTART_INSTANCE");
		if (! job_name) {
			nih_fatal (_("missing job name"));
			return 1;
		}

		if (us_inst) {
			char *env_str;
			NIH_MUST (env_str = nih_sprintf (NULL,
				"UPSTART_INSTANCE=%s", us_inst));
			env = nih_str_array_add (&env, NULL, NULL, us_inst);
		}
	}

	if (control_get_job_by_name (proxy, job_name, &job_path)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to stop job: %s"),
			   err->message);

		return 1;
	}

	if (! job_path) {
		nih_fatal (_("No such job: %s"), job_name);
		return 1;
	}

	NIH_MUST (job = nih_dbus_proxy_new (NULL, conn,
			"com.ubuntu.Upstart.Job", job_path));

	if (job_class_stop (job, env)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to stop job: %s"),
			   err->message);

		return 1;
	}

	return 0;
}

/**
 * status_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "status" command; it takes one or more
 * arguments that give the names of jobs that should be queried for their
 * current status.
 *
 * Returns: command exit status.
 **/
int
status_action (NihCommand   *command,
	       char * const *args)
{
	NihDBusProxy *job;
	char **jobs;
	char **instances;

	if (control_get_all_jobs (proxy, &jobs)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to get job list: %s"),
			   err->message);

		return 1;
	}

	for (; *jobs; jobs++) {
		int inst_count = 0;
		char *instname;
		char *name = rindex (*jobs, '/') + 1;

		if (args[0] && fnmatch (args[0], name, 0))
			continue;

		NIH_MUST (job = nih_dbus_proxy_new (NULL, conn,
					"com.ubuntu.Upstart.Jobs", *jobs));

		if (job_class_get_all_instances (job, &instances)) {
			NihError *err;

			err = nih_error_get ();
			nih_fatal (_("Unable to get instance list: %s"),
				   err->message);

			return 1;
		}

		for (; instances[inst_count]; inst_count++);

		if (! inst_count) {
			printf ("%s not running\n", name);
			continue;
		}

		if (inst_count > 1) {
			printf ("%s %d instances\n", name, inst_count);
			continue;
		}

		instname = instances[0] + strlen (*jobs) + 1;

		if (! strncmp ("_", instname, 2)) {
			printf ("%s running\n", name);
		} else {
			printf ("%s 1 instance\n", name);
		}

	}
	return 0;
}

/**
 * list_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "list" command; it takes an optional
 * pattern argument and returns a list of jobs matching that argument, or
 * all jobs if the argument is not given.
 *
 * Returns: command exit status.
 **/
int
list_action (NihCommand   *command,
	     char * const *args)
{
	char **list;

	if (control_get_all_jobs (proxy, &list)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to get job list: %s"),
			   err->message);

		return 1;
	}

	for (; *list; list++) {
		char *name = *list + strlen ("/com/ubuntu/Upstart/jobs/");

		if (args[0] && fnmatch (args[0], name, 0))
			continue;
		printf ("%s\n", name);
	}

	return 0;
}

/**
 * emit_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "emit" command; it takes an event name
 * followed by environment variables.
 *
 * It causes that event to be emitted, and outputs job status changes caused
 * by the event until the event has been handled, unless no_wait is set.
 *
 * Returns: command exit status.
 **/
int
emit_action (NihCommand   *command,
	     char * const *args)
{
	if (! args || ! *args) {
		nih_fatal (_("missing event name"));
		return 1;
	}

	if (control_emit_event (proxy, args[0], args + 1)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to emit event: %s"),
			   err->message);

		return 1;
	}

	return 0;
}

/**
 * jobs_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "jobs" command.  It takes no arguments,
 * and subscribes itself to notification of job status changes; outputting
 * each one until the command is terminated.
 *
 * Returns: command exit status.
 **/
int
jobs_action (NihCommand   *command,
	     char * const *args)
{
	nih_fatal ("Not yet implemented");
	exit (1);
}

/**
 * version_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "version" command; it requests the
 * package version string from the init daemon and outputs the response.
 *
 * Returns: command exit status.
 **/
int
version_action (NihCommand   *command,
		char * const *args)
{
	char *ver;

	if (control_get_version (proxy, &ver)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to get version: %s"),
			   err->message);

		return 1;
	}

	printf ("%s\n", ver);
	return 0;
}

/**
 * log_priority_action:
 * @command: NihCommand invoked,
 * @args: command-line arguments.
 *
 * This function is called for the "log-priority" command; it takes a string
 * log priority which it converts into a NihLogLevel to be sent to the server.
 * No response is expected.
 *
 * Returns: command exit status.
 **/
int
log_priority_action (NihCommand   *command,
		     char * const *args)
{
	int priority;

	if (! strcmp (args[0], "debug")) {
		priority = NIH_LOG_DEBUG;
	} else if (! strcmp (args[0], "info")) {
		priority = NIH_LOG_INFO;
	} else if (! strcmp (args[0], "message")) {
		priority = NIH_LOG_MESSAGE;
	} else if (! strcmp (args[0], "warn")) {
		priority = NIH_LOG_WARN;
	} else if (! strcmp (args[0], "error")) {
		priority = NIH_LOG_ERROR;
	} else if (! strcmp (args[0], "fatal")) {
		priority = NIH_LOG_FATAL;
	} else {
		fprintf (stderr, _("%s: invalid priority\n"), program_name);
		nih_main_suggest_help ();
		return 1;
	}

	if (control_set_log_prio (proxy, priority)) {
		NihError *err;

		err = nih_error_get ();
		nih_fatal (_("Unable to set log priority: %s"),
			   err->message);

		return 1;
	}
}


#ifndef TEST
/**
 * options:
 *
 * Command-line options accepted for all arguments.
 **/
static NihOption options[] = {
	NIH_OPTION_LAST
};


/**
 * start_options:
 *
 * Command-line options accepted for the start command.
 **/
NihOption start_options[] = {
	{ 'n', "no-wait", N_("do not wait for job to start before exiting"),
	  NULL, NULL, &no_wait, NULL },

	NIH_OPTION_LAST
};

/**
 * stop_options:
 *
 * Command-line options accepted for the stop command.
 **/
NihOption stop_options[] = {
	{ 'n', "no-wait", N_("do not wait for job to stop before exiting"),
	  NULL, NULL, &no_wait, NULL },

	NIH_OPTION_LAST
};

/**
 * status_options:
 *
 * Command-line options accepted for the status command.
 **/
NihOption status_options[] = {
	NIH_OPTION_LAST
};

/**
 * list_options:
 *
 * Command-line options accepted for the list command.
 **/
NihOption list_options[] = {
	NIH_OPTION_LAST
};

/**
 * emit_options:
 *
 * Command-line options accepted for the emit command.
 **/
NihOption emit_options[] = {
	{ 'n', "no-wait", N_("do not wait for event to finish before exiting"),
	  NULL, NULL, &no_wait, NULL },
	{ 'e', NULL, N_("set environment variable in jobs changed by this event"),
	  NULL, "NAME[=VALUE]", &emit_env, env_option },

	NIH_OPTION_LAST
};

/**
 * jobs_options:
 *
 * Command-line options accepted for the jobs command.
 **/
NihOption jobs_options[] = {
	NIH_OPTION_LAST
};

/**
 * version_options:
 *
 * Command-line options accepted for the version command.
 **/
NihOption version_options[] = {
	NIH_OPTION_LAST
};

/**
 * log_priority_options:
 *
 * Command-line options accepted for the log-priority command.
 **/
NihOption log_priority_options[] = {
	NIH_OPTION_LAST
};


/**
 * job_group:
 *
 * Group of commands related to jobs.
 **/
static NihCommandGroup job_commands = { N_("Job") };

/**
 * event_group:
 *
 * Group of commands related to events.
 **/
static NihCommandGroup event_commands = { N_("Event") };

/**
 * commands:
 *
 * Commands accepts as the first non-option argument, or program name.
 **/
static NihCommand commands[] = {
	{ "start", N_("JOB [ENV]"),
	  N_("Start jobs."),
	  N_("JOB is a job to start. ENV may be zero or more variables to "
	     "place in the job's environment."),
	  &job_commands, start_options, start_action },

	{ "stop", N_("JOB [KEY=VALUE...]"),
	  N_("Stop jobs."),
	  N_("JOB is a job to stop. ENV may be zero or more variables to "
	     "place in the job's environment."),
	  &job_commands, stop_options, stop_action },

	{ "status", N_("[PATTERN]"),
	  N_("Query status of jobs."),
	  N_("PATTERN is a pattern to select which jobs to display by name"),
	  &job_commands, status_options, status_action },

	{ "list", N_("[PATTERN]"),
	  N_("List known jobs."),
	  N_("PATTERN is a pattern to select which jobs to display by name"),
	  &job_commands, list_options, list_action },

	{ "emit", N_("EVENT [ENV]..."),
	  N_("Emit an event."),
	  N_("EVENT is the name of an event the init daemon should emit, "
	     "which may have zero or more environment variables specified by "
	     "ARG.  These may be matched in the job definition, and are passed "
	     "to any scripts run by the job."),
	  &event_commands, emit_options, emit_action },

	{ "jobs", NULL,
	  N_("Receive notification of job state changes."),
	  NULL,
	  &job_commands, jobs_options, jobs_action },

	{ "version", NULL,
	  N_("Request the version of the init daemon."),
	  NULL,
	  NULL, version_options, version_action },

	{ "log-priority", N_("PRIORITY"),
	  N_("Change the minimum priority of log messages from the init "
	     "daemon"),
	  N_("PRIORITY may be one of "
	     "`debug' (messages useful for debugging upstart are logged, "
	     "equivalent to --debug on kernel command-line); "
	     "`info' (messages about job goal and state changes, as well "
	     "as event emissions are logged, equivalent to --verbose on the "
	     "kernel command-line); "
	     "`message' (informational and debugging messages are suppressed, "
	     "the default); "
	     "`warn' (ordinary messages are suppressed whilst still "
	     "logging warnings and errors); "
	     "`error' (only errors are logged, equivalent to --quiet on "
	     "the kernel command-line) or "
	     "`fatal' (only fatal errors are logged)."),
	  NULL, log_priority_options, log_priority_action },

	NIH_COMMAND_LAST
};


int
main (int   argc,
      char *argv[])
{
	int ret;

	nih_main_init (argv[0]);

	/* Connect to the daemon */
	conn = nih_dbus_connect ("unix:abstract=/com/ubuntu/upstart", NULL);
	if (! conn) {
		NihError *err;

		err = nih_error_get ();

		nih_fatal (_("Unable to establish control connection: %s"),
			   err->message);

		exit (1);
	}

	NIH_MUST (proxy = nih_dbus_proxy_new (NULL, conn, "com.ubuntu.Upstart",
			"/com/ubuntu/Upstart"));

	ret = nih_command_parser (NULL, argc, argv, options, commands);
	if (ret < 0)
		exit (1);

	return ret;
}
#endif
